#version 330 core

out vec4 LFragment; 

in vec3 Normal;
in vec3 FragPos;
in mat3 TBN;
in vec3 theTexCoord;
in vec3 Position;

uniform vec3 cameraPosition;
uniform samplerCube skybox; 

uniform bool flatshading = true;
uniform vec3 fogcolour = vec3(1.0, 0.9, 1.0);
uniform float fogdistance = 10;
uniform float fogheight = 0.9;

float gamma = 2.2;

struct Material {
    sampler2D texture_diffuse1;
    sampler2D texture_diffuse2;
    sampler2D texture_diffuse3;
    sampler2D texture_specular1;
    sampler2D texture_specular2;
    sampler2D texture_specular3;
    sampler2D texture_normal1;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
    float shininess;
};

#define DIRECTIONAL_LIGHT 0
#define POINT_LIGHT 1
#define SPOT_LIGHT 2


        /***** Light Attenuation 
        modelShader->setVec3("material.specular", glm::vec3(0.7));
        modelShader->setVec3("material.diffuse", glm::vec3(1.0, 0.7, 0.8));
        modelShader->setVec3("material.ambient", glm::vec3(0.2));
        modelShader->setFloat("material.shininess", 32);

         *
         * Constant x, Linear y, Quadratic z
         *
         *  Distance   Constant 	Linear 	  Quadratic
         *      7 	    1.0 	     0.7        1.8   
         *      13 	    1.0 	     0.35       0.44  
         *      20 	    1.0 	     0.22       0.20  
         *      32 	    1.0 	     0.14       0.07  
         *      50 	    1.0 	     0.09       0.032 
         *      65 	    1.0 	     0.07       0.017 
         *      100 	1.0 		 0.045      0.0075  
         *      160 	1.0 		 0.027      0.0028  
         *      200 	1.0 		 0.022      0.0019  
         *      325 	1.0 		 0.014      0.0007  
         *      600 	1.0 		 0.007      0.0002  
         *      3250 	1.0 	 	 0.0014     0.000007
         *
         */




struct Light {
    int type;
    vec3 direction;
    vec3 position;
    float angle_cutoff;
    float outer_angle_cutoff;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
    bool attenuate;
    vec3 attenuation;
};

#define MAX_LIGHTS 10

uniform Light lights[MAX_LIGHTS];
uniform int num_lights;

uniform Material material;

vec4 
reflection(vec3 norm) 
{
    vec3 reflectiveness = material.specular * material.shininess / 50;
    vec3 I = normalize(FragPos - cameraPosition);
    vec3 R = reflect(I, norm);
    return texture(skybox, R) * (material.shininess / 90);
}

float
spotlight(int i, vec3 lightDir) 
{

    float theta = dot(lightDir, normalize(-lights[i].direction));
    float epsilon = lights[i].angle_cutoff - lights[i].outer_angle_cutoff;
    float intensity = clamp((theta - lights[i].outer_angle_cutoff)/epsilon, 0.0, 1.0);

    return intensity;
}

float 
attenuation(int i) 
{
    float dist = length(lights[i].position - FragPos);
    // for no gamma correction
    //float attenuation = 1.0 / (lights[i].attenuation.x + lights[i].attenuation.y *
    //dist + lights[i].attenuation.y * (dist * dist));

    // for gamma correction 
    float attenuation = 1.0 / (lights[i].attenuation.x + lights[i].attenuation.y *
    dist + lights[i].attenuation.y * (dist));
    return attenuation;
}

vec3
specular(int i, vec3 lightDir, vec3 norm)
{
    float spec = max(dot(normalize(cameraPosition - FragPos), reflect(-lightDir, norm)), 0.0);
    spec = pow(spec, material.shininess);
    vec3 specmap = vec3(0);
    if (!flatshading) {
        specmap = vec3(texture(material.texture_specular1, vec2(theTexCoord)));
    }
    vec3 specular = lights[i].specular * (spec * specmap);
    return specular;
}

vec3
diffuse(int i, vec3 lightDir, vec3 norm)
{
    float diff_val = max(dot(norm, lightDir), 0.0);
    vec3 diffuse = lights[i].diffuse * diff_val;
    return diffuse;
}


void 
main() 
{ 
    vec3 endResult = vec3(0);
    // normal
    vec3 norm = vec3(1,1,1);
    if (!flatshading) {
        norm = texture(material.texture_normal1, vec2(theTexCoord)).rgb;
    }
    norm = norm * 2.0 - 1.0;
    norm = normalize(TBN * norm);


    for (int i = 0; i < num_lights; i++) {

        vec3 ambient = lights[i].ambient * material.ambient;


        // diffuse
        vec3 lightDir = vec3(0);
        if (lights[i].type == POINT_LIGHT || lights[i].type == SPOT_LIGHT) {
            lightDir = normalize(lights[i].position - FragPos);
        } else if (lights[i].type == DIRECTIONAL_LIGHT) {
            lightDir = normalize(-lights[i].direction);
        } 

        vec3 diffuse = diffuse(i, lightDir, norm);

        vec3 specular = specular(i, lightDir, norm);

        vec3 result = ambient;
        float intensity = 1.0;

        if (lights[i].type == SPOT_LIGHT) {
            intensity = spotlight(i, lightDir);
        }

        result = ambient + (diffuse + specular) * intensity;

        if (lights[i].attenuate) {
            result *= attenuation(i);
        }

        endResult += result;
    }

    vec3 dist = cameraPosition - FragPos;
    vec3 entrydist = (dist / dist.y) * (cameraPosition.y - fogheight);


    if (cameraPosition.y > fogheight) {
        dist -= entrydist;
    } else {
        // not intersect 
        if (FragPos.y > fogheight) {
        // intersect
            dist =  entrydist;
        }
        
    }

    float fogfactor =  length(dist) * length(dist)  / (fogdistance * fogdistance);
//    fogfactor *= min(0.0, entrydist.y) / (min(entrydist.y, 1.0));
    if (FragPos.y > fogheight && cameraPosition.y > fogheight) {
        fogfactor = 0;
    }


    vec4 parta = (vec4(endResult, 1.0)) * texture(material.texture_diffuse1, vec2(theTexCoord)); 
    parta.rgb += fogfactor * fogcolour;



    LFragment = parta;

     

//   LFragment.rgb = parta.rgb;
//   LFragment.rbg += ((1.0 - parta.a) * skyval.rgb);

    // gamma correction
//    LFragment.rgb = pow(parta.rgb, vec3(1.0/gamma));
//    LFragment = vec4(norm, 1);
    
}

